// BKbin2obj.cpp:      .
//

#include "pch.h"
#include "getopt.h"
#include "BKbin2obj.h"
#include "Globals.h"
#include "LabelManager.h"
#include "Object.h"

#include <clocale>
#pragma warning(disable:4996)

std::wstring strType, strLabel, strFinalLabel, strInFileName, strOutFileName;
/*
:
bColor -  /  
bTransparency - 
*/
bool bColor = false, bTransparency = false, bEven = false;
int nEvenBound = 0;

int wmain(int argc, wchar_t *argv[], wchar_t *envp[])
{
	_wsetlocale(LC_ALL, L"Russian");
	InitGlobalParameters(&g_GlobalParameters);
	static struct option long_options[] =
	{
		{ L"help",         ARG_NONE, nullptr, L'?' },
		{ L"source",       ARG_REQ,  nullptr, L's' },
		{ L"even",         ARG_OPT,  nullptr, L'e' },
		{ L"label",        ARG_REQ,  nullptr, L'l' },
		{ L"final",        ARG_REQ,  nullptr, L'f' },
		{ L"color",        ARG_NONE, nullptr, L'c' },
		{ L"transparency", ARG_NONE, nullptr, L't' },
		{ nullptr,         ARG_NULL, nullptr, ARG_NULL }
	};
	static wchar_t optstring[] = L"?s:e::l:f:ct";
	bool bShowHelp = false;
	int option_index = 0;
	int c;

	while ((c = getopt_long(argc, argv, optstring, long_options, &option_index)) != -1)
	{
		// Handle options
		c = tolower(c);

		switch (c)
		{
			case L'?':
				bShowHelp = true;
				break;

			case L's':
				if (optarg)
				{
					strType = std::wstring(optarg);
					strType = strType.substr(0, 3);
					strType[0] = toupper(strType[0]);
					strType[1] = toupper(strType[1]);
					strType[2] = toupper(strType[2]);
				}

				break;

			case L'e':
				bEven = true;

				if (optarg)
				{
					int nTmp = _tcstol(optarg, nullptr, 8) & 0xffff;

					if (0 <= nTmp && nTmp <= 0100000)
					{
						nEvenBound = nTmp;
					}
				}

				break;

			case L'l':
				bEven = true;

				if (optarg)
				{
					strLabel = std::wstring(optarg);

					if (!strLabel.empty())
					{
						for (wchar_t &n : strLabel)
						{
							n = toupper(n);
						}
					}
				}

				break;

			case L'f':
				bEven = true;

				if (optarg)
				{
					strFinalLabel = std::wstring(optarg);

					if (!strFinalLabel.empty())
					{
						for (wchar_t &n : strFinalLabel)
						{
							n = toupper(n);
						}
					}
				}

				break;

			case L'c':
				bColor = true;
				break;

			case L't':
				bTransparency = true;
				break;
		}

		if (bShowHelp)
		{
			Usage();
			return 0;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc < 1)
	{
		wprintf(L"   :     .\n\n");
		Usage();
	}
	else
	{
		strInFileName = std::wstring(*argv);
		argc--;
		argv++;

		if (argc >= 1)
		{
			strOutFileName = std::wstring(*argv);
		}
		else
		{
			strOutFileName = strInFileName;
		}

		if (strLabel.empty())
		{
			//      
			wchar_t pName[_MAX_FNAME] = { 0 };
			_wsplitpath_s(strInFileName.c_str(), nullptr, 0, nullptr, 0, pName, _MAX_FNAME, nullptr, 0);
			//     
			strLabel = std::wstring(pName);
		}

		g_GlobalParameters.nModeLinkage = -1;

		if (strType == L"BIN")
		{
			ConvertBin();
		}
		else if (strType == L"IMG")
		{
			ConvertImage();
		}
	}

	return 0;
}

void ConvertImage()
{
	CImage img;

	if (SUCCEEDED(img.Load(strInFileName.c_str())))
	{
		int w = img.GetWidth();
		int h = img.GetHeight();
		//        .
		int size = 0;

		if (bColor)
		{
			//   .
			// 1  - 2 
			size = ((w + 3) / 4) * h * (bTransparency ? 2 : 1);
		}
		else
		{
			//  - .
			// 1  - 1 
			size = ((w + 7) / 8) * h * (bTransparency ? 2 : 1);
		}

		if (size < 65536)
		{
			if (size >= 32768)
			{
				wprintf(L"!       .\n");
			}

			// 
			if (bColor)
			{
				convertBitmap2(&g_Memory.b[BASE_ADDRESS], &img, bTransparency);
			}
			else
			{
				convertBitmap1(&g_Memory.b[BASE_ADDRESS], &img, bTransparency);
			}

			CBKToken token(strLabel); //  
			g_labelGlobalDefs.AddLabel(&token, BASE_ADDRESS, DEFINITE_LABEL);

			if (bEven)
			{
				if (nEvenBound)
				{
					//      2
					size = size ? ((size - 1) | (nEvenBound - 1)) + 1 : 0;
				}
				else
				{
					//    
					size++;
					size &= ~1;
				}
			}

			g_GlobalParameters.nProgramLength = size;
			MakeObj(strOutFileName, std::wstring(L".obj"));
		}
		else
		{
			wprintf(L"   .\n");
		}

		img.Destroy();
	}
	else
	{
		wprintf(L"    %s\n", strInFileName.c_str());
	}
}


//     
void ConvertBin()
{
	FILE *inf = _wfopen(strInFileName.c_str(), L"rb");

	if (inf)
	{
		//   
		fseek(inf, 0, SEEK_END);
		long size = ftell(inf);
		fseek(inf, 0, SEEK_SET);

		if (size < 65536)
		{
			if (size >= 32768)
			{
				wprintf(L"!       .\n");
			}

			fread(&g_Memory.b[BASE_ADDRESS], 1, size, inf); //  
			CBKToken token(strLabel); //  
			g_labelGlobalDefs.AddLabel(&token, BASE_ADDRESS, DEFINITE_LABEL);

			if (bEven)
			{
				if (nEvenBound)
				{
					//      2
					size = size ? ((size - 1) | (nEvenBound - 1)) + 1 : 0;
				}
				else
				{
					//    
					size++;
					size &= ~1;
				}
			}

			if (!strFinalLabel.empty())
			{
				CBKToken finalToken(strFinalLabel); //  
				g_labelGlobalDefs.AddLabel(&finalToken, BASE_ADDRESS + size, DEFINITE_LABEL);
			}

			g_GlobalParameters.nProgramLength = size;
			MakeObj(strOutFileName, std::wstring(L".obj"));
		}
		else
		{
			wprintf(L"   .\n");
		}

		fclose(inf);
	}
	else
	{
		wprintf(L"   %s\n", strInFileName.c_str());
	}
}

//   pdp11asm vinxru
//   
void convertBitmap2(uint8_t *destp, CImage *img, bool t)
{
	int w = img->GetWidth();
	int h = img->GetHeight();
	int destbpl = (w + 3) / 4 * (t ? 2 : 1);

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			COLORREF olor = img->GetPixel(x, y);
			//  -  .
			bool tr = (olor == 0xFF00FF);

//          if (olor == 0)
//          {
//              olor = 0;
//          }
//          else if (olor == 0x0000FF)
//          {
//              olor = 1;
//          }
//          else if (olor == 0x00FF00)
//          {
//              olor = 2;
//          }
//          else
//          {
//              olor = 3;
//          }

			//  ,    
			if ((olor & 0xff0000) && !(olor & 0x00f8f8)) //     - ,     
			{
				olor = 1; //    
			}
			else if ((olor & 0x00ff00) && !(olor & 0xf800f8)) //     - ,     
			{
				olor = 2; //    
			}
			else if ((olor & 0x0000ff) && !(olor & 0xf8f800)) //     - ,     
			{
				olor = 3; //    
			}
			else
			{
				//    -  .  .
				olor = 0; // 
			}

			if (t)
			{
				if (!tr)
				{
					if (w >= 16)
					{
						destp[x / 4 + x / 4 / 2 * 2 + 2 + y * destbpl] |= olor << ((x & 3) * 2);
						destp[x / 4 + x / 4 / 2 * 2 + y * destbpl] |= 3 << ((x & 3) * 2);
					}
					else
					{
						destp[x / 4 * 2 + 1 + y * destbpl] |= olor << ((x & 3) * 2);
						destp[x / 4 * 2 + y * destbpl] |= 3 << ((x & 3) * 2);
					}
				}
			}
			else
			{
				destp[x / 4 + y * destbpl] |= (olor << ((x & 3) * 2));
			}
		}
	}
}

// -----------------------------------------------------------------------------
//    
void convertBitmap1(uint8_t *destp, CImage *img, bool t)
{
	int w = img->GetWidth();
	int h = img->GetHeight();
	size_t destbpl = (w + 7) / 8 * (t ? 2 : 1);

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			COLORREF c = img->GetPixel(x, y);
			bool tr = (c == 0xFF00FF);

			if (c == 0)
			{
				c = 0;
			}
			else
			{
				c = 1;
			}

			if (t)
			{
				if (!tr)
				{
					if (w >= 16)
					{
						if (c)
						{
							destp[x / 8 + x / 8 / 2 * 2 + 2 + y * destbpl] |= 1 << (x & 7);
						}

						destp[x / 8 + x / 8 / 2 * 2 + y * destbpl] |= 1 << (x & 7);
					}
					else
					{
						if (c)
						{
							destp[x / 8 * 2 + 1 + y * destbpl] |= 1 << (x & 7);
						}

						destp[x / 8 * 2 + y * destbpl] |= 1 << (x & 7);
					}
				}
			}
			else
			{
				if (c)
				{
					destp[x / 8 + y * destbpl] |= c << (x & 7);
				}
			}
		}
	}
}

void Usage()
{
	wprintf(L"        Turbo8.\n" \
	        L"() 2014-2021 gid\n\n" \
	        L":\n" \
	        L"BKbin2obj -? (--help)\n" \
	        L"    .\n\n" \
	        L"BKbin2obj [-s<type>][-c][-t][-e[0bound]][-l<label_name>][-f<label_name>] <input_file_name> [output_file_name]\n" \
	        L"  -s<source> (--source <source>) -   .\n" \
	        L"     :\n" \
	        L"    bin -   ;\n" \
	        L"    img -    BMP, GIF, JPEG, PNG  TIFF.\n\n" \
	        L"  -e[0bound] (--even [0bound]) -      .\n" \
	        L"        -     .\n" \
	        L"     -    . ,   - \n" \
	        L"    .    ,     , \n" \
	        L"    .\n\n" \
	        L"  -l<label_name> (--label <label_name>) -   .    \n" \
	        L"    ,      .\n\n" \
	        L"  -f<label_name> (--final <label_name>) -      .\n" \
	        L"       -e,      .\n\n" \
	        L"  ,      img:\n" \
	        L"  -c (--color) -     ,  - -\n" \
	        L"    .\n\n" \
	        L"  -t (--transparency) -  .\n" \
	        L"         pdp11asm vinxru,   ,\n" \
	        L"          .\n\n" \
	        L"  input_file_name -  .\n" \
	        L"  output_file_name -    ,   \n" \
	        L"    ,   .\n");
}

/*
      ,    
 .
*/
