#include "pch.h"
#include "LabelTable.h"


CLabelTable::CLabelTable()
    = default;


CLabelTable::~CLabelTable()
{
	Clear();
}

void CLabelTable::Clear()
{
	LabelTable.clear();
}

// : true -  .
//      false -   ,     
bool CLabelTable::AddLabel(CBKToken *token, int value, uint32_t lt)
{
	if (SearchLabel(token) == -1) //     
	{
		// 
		Label lbl(lt, value, token);
		LabelTable.push_back(lbl);
		return true;
	}

	return false; //     -   
}

/*
   
*/
CBKToken *CLabelTable::GetLabel(size_t n)
{
	if (n < getSize()) //     
	{
		return &LabelTable.at(n).label; // 
	}

	return nullptr; //  null
}

/*
   
*/
void CLabelTable::DeleteLabel(size_t n)
{
	if (n < getSize()) //     
	{
		LabelTable.erase(LabelTable.begin() + n); // 
	}
}

/*
   
*/
void CLabelTable::DeleteLabel(CBKToken *token)
{
	int n = SearchLabel(token); //  

	if (n >= 0) //  
	{
		DeleteLabel(n); //   
	}
}

/*
    ,   ,  -1
 -  .
*/
int CLabelTable::SearchLabel(CBKToken *token)
{
	size_t sz = getSize();

	for (size_t i = 0; i < sz; ++i)
	{
		if (LabelTable.at(i).label.getHash() == token->getHash())
		{
#ifdef _DEBUG

			if (LabelTable.at(i).label.getName() != token->getName())
			{
				wprintf(L"HASH Error: %s:%#zX ~~ %s:%#zX\n", LabelTable.at(i).label.getName().c_str(), LabelTable.at(i).label.getHash(), token->getName().c_str(), token->getHash());
				assert(false);
			}

#endif
			return static_cast<int>(i);
		}
	}

	return -1; //   
}

/*
    
*/
int CLabelTable::GetValue(CBKToken *token)
{
	int n = SearchLabel(token); //  

	if (n >= 0) //  
	{
		return LabelTable.at(n).label_value; //   
	}

	return -1; //   
}

/*
    
*/
int CLabelTable::GetValue(size_t n)
{
	if (n < getSize()) //     
	{
		return LabelTable.at(n).label_value; //   
	}

	return -1; //   
}

/*
    
*/
uint32_t CLabelTable::GetType(CBKToken *token)
{
	int n = SearchLabel(token);

	if (n >= 0)
	{
		return LabelTable.at(n).label_type;
	}

	return UNKNOWN_LABEL;
}

/*
    
*/
uint32_t CLabelTable::GetType(size_t n)
{
	if (n < getSize())
	{
		return LabelTable.at(n).label_type;
	}

	return UNKNOWN_LABEL;
}



CLabelTableRefs::CLabelTableRefs()
{
}


CLabelTableRefs::~CLabelTableRefs()
{
	Clear();
}

// : true -  .
//          .
bool CLabelTableRefs::AddLabel(CBKToken *token, int value, uint32_t lt)
{
	// 
	Label lbl(lt, value, token);
	LabelTable.push_back(lbl);
	return true; //   
}
