#include "pch.h"
#include "Listing.h"
#include "Globals.h"
#include "Reader.h"

std::vector <ListingLine> g_Listing;
#pragma warning(disable:4996)

int offsetpc = g_GlobalParameters.nStartAddress;

//  
void Lst_PrepareLine(int cp)
{
	ListingLine ll;
	ll.nLineNum = g_pReader->getLineNum();
	ll.line = g_pReader->getCurString();
	ListingCmdType lct;
	lct.nPC = cp;
	lct.LT = ListType::LT_BLANKLINE;
	lct.nEndAdr = 0;
	ll.vCMD.push_back(lct);
	g_Listing.push_back(ll);
}

//     
void Lst_AddPrepareLine(int cp, ListType lt)
{
	auto pll = g_Listing.end() - 1;
	auto plct = pll->vCMD.end() - 1;

	if (plct->LT == ListType::LT_BLANKLINE)   //    -  ,      -
	{
		plct->nPC = cp;
		plct->LT = lt;  //   
	}
	else if (plct->LT == ListType::LT_LABEL)  //    - ,    ,    ,
	{
		//       .
		plct->nPC = cp;
		plct->LT = lt;  //   
	}
	else if (lt != ListType::LT_COMMENTARY) //   -   ,    .
	{
		ListingCmdType lct;
		lct.nPC = cp;
		lct.LT = lt;
		lct.nEndAdr = 0;
		pll->vCMD.push_back(lct);
	}
}

//     
void Lst_AddPrepareLine2(int len)
{
	auto pll = g_Listing.end() - 1;
	auto plct = pll->vCMD.end() - 1;
	plct->nEndAdr = len & 0xffff;
}

/*
   4 
1 -  
2 -  
3 -  
4 -  
*/


//   
void Lst_OutLineNum(FILE *of, int nOut, int num)
{
	if (nOut == 0)
	{
		fwprintf(of, L"% 5d ", num);
	}
	else
	{
		fwprintf(of, L"      ");
	}
}
//  
void Lst_OutAddress(FILE *of, int nOutLN, bool bOut, int num, int addr)
{
	Lst_OutLineNum(of, nOutLN, num);

	if (bOut)
	{
		fwprintf(of, L"%07o: ", ((addr - BASE_ADDRESS + offsetpc) & 0xffff));
	}
	else
	{
		fwprintf(of, L"         ");
	}
}
//    
// : bgn_addr -   
//      end_addr -   
void Lst_OutWords(FILE *of, int nOutLN, bool bOutAddr, bool bOut, int num, int &bgn_addr, int end_addr)
{
	Lst_OutAddress(of, nOutLN, bOutAddr, num, bgn_addr);
	int npc = bgn_addr / 2;
	//  .    1  3  
	int nlen = (bOut) ? (end_addr - bgn_addr) / 2 : 0;
	int nmax = 3 - nlen;

	for (int i = 0; i < nlen; ++i)
	{
		fwprintf(of, L"%07o ", g_Memory.w[npc++]);
	}

	for (int i = 0; i < nmax; ++i)
	{
		fwprintf(of, L"        ");
	}

	fwprintf(of, L" ");
	bgn_addr = npc * 2;
}
//    
// : bgn_addr -   
//      end_addr -   
void Lst_OutBytes(FILE *of, int nOutLN, bool bOutAddr, bool bOut, int num, int &bgn_addr, int end_addr)
{
	Lst_OutAddress(of, nOutLN, bOutAddr, num, bgn_addr);
	int npc = bgn_addr;
	//  .    1  5  
	int nlen = (bOut) ? (end_addr - bgn_addr) : 0;
	int nmax = 5 - nlen;

	for (int i = 0; i < nlen; ++i)
	{
		fwprintf(of, L"%04o ", g_Memory.b[npc++]);
	}

	for (int i = 0; i < nmax; ++i)
	{
		fwprintf(of, L"     ");
	}

	bgn_addr = npc;
}

void Lst_OutLineString(FILE *of, int nCnt, std::wstring *str = nullptr)
{
	if (nCnt == 0)
	{
		if (str)
		{
			size_t nLen = str->length();

			if (nLen)
			{
				auto *pBuff = new uint8_t[nLen + 1];

				if (pBuff)
				{
					UNICODEtoBK(*str, pBuff, nLen, false);
					pBuff[nLen] = 0;
					fprintf(of, "%s", pBuff);
					delete[] pBuff;
					return;
				}
			}
		}
	}

	fwprintf(of, L"\n");
}


void MakeListing(const std::wstring &strName, const std::wstring &strExt)
{
	offsetpc = g_GlobalParameters.nStartAddress;
	wchar_t pDrive[_MAX_DIR] = { 0 };
	wchar_t pPath[_MAX_PATH] = { 0 };
	wchar_t pName[_MAX_FNAME] = { 0 };
	wchar_t pExt[_MAX_EXT] = { 0 };
	_wsplitpath_s(strName.c_str(), pDrive, _MAX_DIR, pPath, _MAX_PATH, pName, _MAX_FNAME, pExt, _MAX_EXT);
	//    
	std::wstring outName = std::wstring(pDrive) + std::wstring(pPath) + std::wstring(pName) + strExt;
	FILE *of = _wfopen(outName.c_str(), L"wt");

	if (of == nullptr)
	{
		wprintf(L"   %s\n", outName.c_str());
	}

	for (auto &ll : g_Listing)
	{
		int nCount = 0;

//    ,  
		for (auto &err : ll.errors)
		{
			size_t len = err.length();
			auto *pbuf = new uint8_t[len + 1];

			if (pbuf)
			{
				UNICODEtoBK(err, pbuf, len, false);
				pbuf[len] = 0;
				fprintf(of, "%s", pbuf);
				delete[] pbuf;
			}
		}

		for (auto &lct : ll.vCMD)
		{
			// ,       
			switch (lct.LT)
			{
				case ListType::LT_BLANKLINE:
					Lst_OutLineNum(of, nCount++, ll.nLineNum);
					fwprintf(of, L"\n");
					break;

				case ListType::LT_COMMENTARY:
				{
					int adr = 0;
					Lst_OutWords(of, nCount, false, false, ll.nLineNum, adr, 0); //    
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_ASSIGN:
				{
					Lst_OutAddress(of, nCount, false, ll.nLineNum, 0); //    
					fwprintf(of, L"%07o                  ", (lct.nEndAdr & 0xffff)); //    
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_LABEL:
				{
					int adr = lct.nPC;
					Lst_OutWords(of, nCount, true, false, ll.nLineNum, adr, lct.nEndAdr); //    
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_INSTRUCTION:
				{
					int adr = lct.nPC;
					Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, lct.nEndAdr); //    
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_PSC_ADDR:
				{
					int adr = lct.nPC;
					Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, lct.nPC + 2);
					Lst_OutLineString(of, nCount++, &ll.line);
					Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, lct.nEndAdr);
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_PSC_LA:
				{
					offsetpc = GetStartAddress(); //     
					int adr = lct.nPC;
					Lst_OutWords(of, nCount, true, false, ll.nLineNum, adr, 0);
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_PSC_BLKW:
				case ListType::LT_PSC_BLKB:
				case ListType::LT_PSC_END:
				{
					int adr = lct.nPC;
					Lst_OutWords(of, nCount, true, false, ll.nLineNum, adr, 0);
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_PSC_PRINT:
				{
					int adr = lct.nPC;
					Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, lct.nPC + 4);
					Lst_OutLineString(of, nCount++, &ll.line);
					Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, lct.nPC + 6);
					Lst_OutLineString(of, nCount++, &ll.line);
					Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, lct.nEndAdr);
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				case ListType::LT_PSC_FLT2:
				case ListType::LT_PSC_FLT4:
				case ListType::LT_PSC_WORD:
				case ListType::LT_PSC_RAD50:
				{
					//      
					int adr = lct.nPC;
					int len = (lct.nEndAdr - adr) / 2;
					int cnt = len / 3;

					if (len % 3)
					{
						cnt++;
					}

					for (int i = 0; i < cnt; ++i)
					{
						int endadr = (adr + 6 < lct.nEndAdr) ? adr + 6 : lct.nEndAdr;
						Lst_OutWords(of, nCount, true, true, ll.nLineNum, adr, endadr); //    
						Lst_OutLineString(of, nCount++, &ll.line);
					}
				}
				break;

				case ListType::LT_PSC_BYTE:
				case ListType::LT_PSC_ASCII:
					//   5   
				{
					int adr = lct.nPC;
					int len = (lct.nEndAdr - adr);
					int cnt = len / 5;

					if (len % 5)
					{
						cnt++;
					}

					for (int i = 0; i < cnt; ++i)
					{
						int endadr = (adr + 5 < lct.nEndAdr) ? adr + 5 : lct.nEndAdr;
						Lst_OutBytes(of, nCount, true, true, ll.nLineNum, adr, endadr); //    
						Lst_OutLineString(of, nCount++, &ll.line);
					}
				}
				break;

				case ListType::LT_PSC_EVEN:
				{
					int adr = lct.nPC;
					bool bOut = ((lct.nEndAdr - lct.nPC) != 0);
					Lst_OutBytes(of, nCount, true, bOut, ll.nLineNum, adr, lct.nEndAdr);
					Lst_OutLineString(of, nCount++, &ll.line);
				}
				break;

				default:
				{
					fwprintf(of, L"!!! Unworked case !!!\n");
				}
			}
		}
	}

	fclose(of);
}


