#include "pch.h"
#include "Parser.h"
#include "Globals.h"
#include "LabelTable.h"
#include "LabelManager.h"
#include "ErrorManager.h"
#pragma warning(disable:4996)

Registers RegNames[] =
{
	{L"R0", 0},
	{L"R1", 1},
	{L"R2", 2},
	{L"R3", 3},
	{L"R4", 4},
	{L"R5", 5},
	{L"R6", 6},
	{L"R7", 7},
	{L"SP", 6},
	{L"PC", 7},
};


const wchar_t RADIX50[050] =
{
	// 000..007
	L' ', L'A', L'B', L'C', L'D', L'E', L'F', L'G',
	// 010..017
	L'H', L'I', L'J', L'K', L'L', L'M', L'N', L'O',
	// 020..027
	L'P', L'Q', L'R', L'S', L'T', L'U', L'V', L'W',
	// 030..037
	L'X', L'Y', L'Z', L'$', L'.', L' ', L'0', L'1',
	// 040..047
	L'2', L'3', L'4', L'5', L'6', L'7', L'8', L'9'
};


/*
  
: cp -  
: result -   
true -  
false -   
*/
bool AriphmParser(int &cp, int &result, wchar_t &ch)
{
	int accum = 0;  // .       
	result = 0;
	g_GlobalParameters.bInvert = false; //    

	for (;;)
	{
		g_pReader->SkipWhitespaces(ch);
		uint32_t lt = RELATIVE_LABEL;

		//   
		if (ch == L'+') //   +
		{
			ch = g_pReader->readChar(true); //   
		}
		else if (ch == L'-') //   -
		{
			ch = g_pReader->readChar(true); //   
			lt |= NEGATIVE_FLAG; //   
		}

		// ,    
		if (ch == L'/') //  /
		{
			ch = g_pReader->readChar(); //   
			lt |= HALFLABEL_FLAG;   //    

			if (g_pReader->getCurrCharType() != CReader::CHAR_TYPE::LETTERS)
			{
				OutError(ERROR_104); //  
				return false; //  .
			}
		}

		if (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::LETTERS)
		{
			CBKToken label;
			g_pReader->readToken(&label, ch);

			if (g_GlobalParameters.nAriphmType == 0)
			{
				//   
				OutError(ERROR_123);
				return false; //  .
			}

			if (g_GlobalParameters.nAriphmType > 0)
			{
				//   .     
				if (g_GlobalParameters.bInvert)
				{
					g_GlobalParameters.bInvert = false;
					lt |= INVERSE_FLAG;
				}

				AddlLabelReference(&label, cp, lt);
				goto next_op;
			}
			else
			{
				//  
				int n = g_labelGlobalDefs.SearchLabel(&label);

				if (n == -1)
				{
					//   .
					OutError(ERROR_122);
					return false; //  .
				}

				result = g_labelGlobalDefs.GetValue(n);

				//   ,      
				if ((g_labelGlobalDefs.GetType(n) & LABEL_DEFINITE_MASK) == DEFINITE_LABEL)
				{
					result = CorrectAddress(result);
				}

				if (g_GlobalParameters.bInvert)
				{
					g_GlobalParameters.bInvert = false;
					result = ~result;
				}

				if (lt & HALFLABEL_FLAG)
				{
					result = result / 2;
				}
			}
		}
		else if (ch == L'\'')
		{
			//   
			g_GlobalParameters.bInString = true;
			ch = g_pReader->readChar();
			g_GlobalParameters.bInString = false;
			result = UNICODEtoBK_ch(ch);

			if (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::LN)
			{
				OutError(ERROR_108); //     .
				return false; //  .
			}

			ch = g_pReader->readChar(); //     ,   
		}
		else if (ch == L'"')
		{
			//   
			g_GlobalParameters.bInString = true;
			ch = g_pReader->readChar();
			g_GlobalParameters.bInString = false;
			result = UNICODEtoBK_ch(ch);

			if (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::LN)
			{
				OutError(ERROR_109); //     .
				return false; //  .
			}

			g_GlobalParameters.bInString = true;
			ch = g_pReader->readChar();
			g_GlobalParameters.bInString = false;
			result |= (UNICODEtoBK_ch(ch)) << 8;

			if (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::LN)
			{
				OutError(ERROR_109); //     .
				return false; //  .
			}

			ch = g_pReader->readChar(); //     ,   
		}
		else if (ch == L'.') //    .
		{
			ch = g_pReader->readChar(); //   

			if (g_GlobalParameters.nAriphmType < 0)
			{
				//     
				result = CorrectAddress(g_GlobalParameters.nPC);
			}
			else
			{
				/*
				  ,    ,    
				   
				*/
				wchar_t buf[32] = { 0 };
				swprintf_s(buf, 32, L"!%06o\0", g_GlobalParameters.nPC);
				CBKToken lbl;
				lbl.setName(std::wstring(buf));
				lt &= ~(LABEL_DEFINITE_MASK | LABEL_REFERENCE_MASK); //   
				AddlLabelReference(&lbl, cp, CURRENTPC_LABEL | lt);
				goto next_op;
			}
		}
		else if (ch == L'^') //    ^
		{
			bool bInvert = false; //   ^

			if (!Macro11NumberParser(result, ch, bInvert))
			{
				OutError(ERROR_105); //     .
				return false; //  .
			}

			//    ^C,   
			if (bInvert)
			{
				if (g_pReader->SkipWhitespaces(ch)) //   
				{
					OutError(ERROR_105); //     .
					return false; //  .
				}

				continue;
			}
		}
		else if (!AdvancedNumberParser(result, ch))
		{
			// !!!  11          .
			//     .
			OutError(ERROR_105); //     .
			return false; //  .
		}

		if (lt & NEGATIVE_FLAG)
		{
			result = -result;
		}

		if (g_GlobalParameters.bInvert)
		{
			result = ~result;
		}

		accum += result;
next_op:
		g_GlobalParameters.bInvert = false; //   
		// ,   ?
		bool bln = g_pReader->SkipWhitespaces(ch);

		if (bln || !(ch == L'+' || ch == L'-'))
		{
			result = accum;
			return true; //  -  .
		}
	}
}

//       
// : cp -  PC
// : result -    
//      true -  
//      false - 
bool trp240(CBKToken *lbl, int &cp, int &result, wchar_t &ch, uint32_t lt)
{
	AddlLabelReference(lbl, cp, lt);
	return trp242(cp, result, ch);
}

//    
// : cp -  PC
// : result -    
//      true -  
//      false - 
bool trp242(int &cp, int &result, wchar_t &ch)
{
	g_pReader->SkipWhitespaces(ch);

	if ((ch == L'+') || (ch == L'-'))
	{
		//  ,  
		return AriphmParser(cp, result, ch);
	}

	result = 0;
	return true;
}

/*
  , ,      
          sob
: result -   
       token -  
false -  .
*/
bool trp202(CBKToken *token, int &result, wchar_t &ch)
{
	if (ch == L'.') //    .
	{
		//   
		result = g_GlobalParameters.nPC;
		ch = g_pReader->readChar(); //   
		return true;
	}

	g_pReader->readToken(token, ch);
	uint32_t lt;
	return SearchLabelInTables(token, result, lt);
}

//       
void SetAddresationRegister(int data)
{
	if (g_GlobalParameters.bOperandType)
	{
		data <<= 6; //  bOperandType,   
	}

	g_Memory.w[g_GlobalParameters.nPC / 2] |= data;
}

int GetAddresationRegister()
{
	int data  = g_Memory.w[g_GlobalParameters.nPC / 2];

	if (g_GlobalParameters.bOperandType)
	{
		data >>= 6; //  bOperandType,   
	}

	return data;
}

//     
bool ReadRegName(wchar_t &ch)
{
	CBKToken reg;
	g_pReader->readToken(&reg, ch);

	if (ParseRegName(&reg))
	{
		return true;
	}

	return false;
}

/*
  
: reg -  
: true -   
       false -  
*/
bool ParseRegName(CBKToken *reg)
{
	for (auto &RegName : RegNames)
	{
		if (RegName.name == reg->getName())
		{
			SetAddresationRegister(RegName.nNum);
			return true;
		}
	}

	return false;
}

//  (Rn)
// :
// 0 -  
// 1 -   
// 2 -   
// 3 -   
int CheckReg(wchar_t &ch)
{
	if (ch != L'(')
	{
		return 2;
	}

	ch = g_pReader->readChar();

	if (!ReadRegName(ch))
	{
		return 1;
	}

	if (ch != L')')
	{
		return 3;
	}

	ch = g_pReader->readChar();
	return 0;
}


//  
bool Operand_analyse(int &cp, wchar_t &ch)
{
	int result = 0;
	g_pReader->SkipWhitespaces(ch); //     
	CReader::CHAR_TYPE ct = g_pReader->getCurrCharType(); //     ?

	if (ct == CReader::CHAR_TYPE::LN) //  
	{
		goto err115; //   
	}
	else if (ct == CReader::CHAR_TYPE::LETTERS) //   -       
	{
		goto l_letters;
	}
	else if (ct == CReader::CHAR_TYPE::DIGITS || ch == L'^') //   -     6,     67
	{
		goto l_digits;
	}
	else
	{
		if (ch == L'@') //    - ,    
		{
			SetAddresationRegister(010); //   
			ch = g_pReader->readChar(); //   
			ct = g_pReader->getCurrCharType();

			if (ct == CReader::CHAR_TYPE::SPACES || ct == CReader::CHAR_TYPE::LN) //       
			{
err103:
				OutError(ERROR_127); //  .
				return false;
			}

			if (ct == CReader::CHAR_TYPE::LETTERS) //     -     
			{
l_letters:
				CBKToken token;
				g_pReader->readToken(&token, ch); //  

				if (ParseRegName(&token)) //    
				{
					cp -= 2; //  
					return true;
				}

				//    -       
				//      
				trp242(cp, result, ch); //    
				int ret = CheckReg(ch); //         

				switch (ret)
				{
					case 1:
						goto err113; //   ,      

					case 2:     //    -   [10]+67
						AddlLabelReference(&token, cp, OFFSET_LABEL);
						goto l_adr67;

					case 3:
						goto err115; //      ,   - 
				}

				AddlLabelReference(&token, cp, RELATIVE_LABEL);
				goto l_adr60; //   -    -  [10]+6
			}

			if (ct == CReader::CHAR_TYPE::DIGITS || ch == L'^') //     - 
			{
l_digits:

				if (!AriphmParser(cp, result, ch)) //   
				{
					return false;
				}

				int ret = CheckReg(ch); //         

				switch (ret)
				{
					case 1:
err113:                 //   ,      
						OutError(ERROR_113); //    .
						return false;

					case 2:
						result -= (cp + 2);
						result = CorrectOffset(result);
l_adr67:                //    -   67
						SetAddresationRegister(067);
						g_Memory.w[cp / 2] += result;
						return true;

					case 3:
err115:                 //      ,   
						OutError(ERROR_127); //   .
						return false;
				}

l_adr60:
				SetAddresationRegister(060); //   -    -  7
				g_Memory.w[cp / 2] += result;
				return true;
			}
		}

		if (ch == L'#') //    - , (    )
		{
			SetAddresationRegister(027); //    
			ch = g_pReader->readChar(); //   

			if (!AriphmParser(cp, result, ch)) //       
			{
				return false;
			}

			g_Memory.w[cp / 2] += result; //      - 
			return true;
		}

		if (ch == L'-') //    - ,        
		{
			ct = g_pReader->getNextCharType(); //    

			if (ct != CReader::CHAR_TYPE::OTHERS || g_pReader->getNextChar() == L'^') //    , ,    
			{
				goto l_digits;      //    
			}

			ch = g_pReader->readChar(); //  -    
			int ret = CheckReg(ch); //     

			switch (ret)
			{
				case 1:
					goto err113; //   ,      

				case 2:
					goto err103; //   .  - 

				case 3:
					goto err115; //      ,   
			}

			SetAddresationRegister(040); //   -  
		}
		else //      
		{
			int ret = CheckReg(ch); //     

			switch (ret)
			{
				case 1:
					goto err113; //   ,      

				case 2:
					goto l_digits; //   .  - 

				case 3:
					goto err115; //      ,   
			}

			//   -  
			if (ch == L'+') //       +
			{
				ch = g_pReader->readChar(); //  
				SetAddresationRegister(020); //    
			}
			else
			{
				//   @(Rn)  
				//  MACRO-11    @0(Rn)
				if (GetAddresationRegister() == 010)
				{
					goto err103; //   
				}

				SetAddresationRegister(010); //   - 
			}
		}

		cp -= 2; //  
	}

	return true;
}

//   .
//   -       - true
//    -  - false    ,  
bool needChar(wchar_t nch, wchar_t &ch)
{
	bool bln = g_pReader->SkipWhitespaces(ch); //   

	if (!bln) //    
	{
		return (ch == nch); // ,   ,   .
	}

	return false;
}


////////////////////////////////////////////////////////////////////////
//  .

bool HexNumberParser(wchar_t &ch, int &ret)
{
	bool bOk = true;
	ret = 0;

	while (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS || g_pReader->getCurrCharType() == CReader::CHAR_TYPE::LETTERS)
	{
		int d = 0;

		if (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS)
		{
			d = (ch - L'0');
		}
		else if (L'A' <= ch && ch <= L'F')
		{
			d = 10 + (ch - L'A');
		}
		else
		{
			bOk = false;
			break;
		}

		ret *= 16;
		ret += d;
		ch = g_pReader->readChar();
	}

	return bOk;
}

bool DecNumberParser(wchar_t &ch, int &ret)
{
	ret = 0;

	while (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS)
	{
		ret *= 10;
		ret += (ch - L'0');
		ch = g_pReader->readChar();
	}

	return true;
}

bool OctNumberParser(wchar_t &ch, int &ret)
{
	ret = 0;
	bool bOk = true;

	while (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS)
	{
		if (L'0' <= ch && ch <= L'7')
		{
			ret *= 8;
			ret += ch & 7;
		}
		else
		{
			//   8  9      ,   
			bOk = false;
			break;
		}

		ch = g_pReader->readChar();
	}

	return bOk;
}

bool BinNumberParser(wchar_t &ch, int &ret)
{
	bool bOk = true;
	ret = 0;

	while (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS)
	{
		if (L'0' <= ch && ch <= L'1')
		{
			ret *= 2;
			ret += ch & 1;
		}
		else
		{
			bOk = false;
			break;
		}

		ch = g_pReader->readChar();
	}

	return bOk;
}


/*    MACRO-11
 Copyright (c) 2001, Richard Krehbiel
     .
  :  15 - ,  8  - ,  7  - 
  :  
   4 :  

     2, 3  4  
 */

/* Parse PDP-11 64-bit floating point format. */
/* Give a pointer to "size" words to receive the result. */
/* Note: there are probably degenerate cases that store incorrect
results.  For example, I think rounding up a FLT2 might cause
exponent overflow.  Sorry. */
/* Note also that the full 49 bits of precision probably aren't
available on the source platform, given the widespread application
of IEEE floating point formats, so expect some differences.  Sorry
again. */
bool parse_float(wchar_t &ch,
                 int size,   // 1  ^f; 2  .FLT2; 4  .FLT4
                 uint16_t *flt) //   1..4 ,   
{
	//       .
	std::wstring str;

	while (g_pReader->getCurrCharType() != CReader::CHAR_TYPE::SPACES && g_pReader->getCurrCharType() != CReader::CHAR_TYPE::LN)
	{
		//       
		if (ch == L'+' || ch == L'-' || ch == L'E' || ch == L'.' || g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS)
		{
			if (ch == L'.')
			{
				ch = L','; //    ,    scanf  ','
			}

			str.push_back(ch); //     .
			ch = g_pReader->readChar();
		}
		else
		{
			break;
		}
	}

	for (int i = 0; i < size; ++i)
	{
		flt[i] = 0;  //  
	}

	double d = 0;   // 
	int n = 0;
	int i = swscanf(str.c_str(), L"%lf%n", &d, &n); //  ,   i == -  

	if (i == 0)
	{
		OutError(ERROR_131); //   
		return false;    //   .
	}

	//    ,   -  
	// -  ,     
	if (size_t(n) < str.length())
	{
		OutError(ERROR_103); //   
	}

	uint16_t uexp = 0;              // 
	uint64_t ufrac = 0;             // 
	uint16_t sign = 0;              //  

	if (d != 0.0)
	{
		int sexp;   //  
		double frac = frexp(d, &sexp);  //    

		if (frac < 0)
		{
			sign = 0100000;             //  
			frac = -frac;               //  
		}

		/*    -  2   56 */
		auto ufrac = uint64_t(frac * 72057594037927936.0);  //   

		/* Round from FLT4 to FLT2 */
		if (size < 4)
		{
			if (size < 2)
			{
				ufrac += 0x800000000000;    //   16- 
			}
			else
			{
				ufrac += 0x80000000;        //   32- 
			}

			if (ufrac >= 0x100000000000000) // ? (    2   56)
			{
				ufrac >>= 1;                // 
				sexp--;
			}
		}

		if (sexp < -128 || sexp > 127)
		{
			//    ?
			OutError(ERROR_131); //   
			return false;
		}

		uint16_t uexp = sexp + 128;     //    - 128
		uexp &= 0xff;                   //    
	}

	//  d == 0.0,    -  0
	flt[0] = uint16_t((sign | (uexp << 7) | ((ufrac >> 48) & 0x7F))); //  

	if (size > 1)
	{
		flt[1] = uint16_t((ufrac >> 32) & 0xffff); //  

		if (size > 2)
		{
			flt[2] = uint16_t((ufrac >> 16) & 0xffff);  // 
			flt[3] = uint16_t(ufrac & 0xffff);          // 
		}
	}

	return true;
}


/*
 
: true -   result 
false -  ,    result 

   ,    

  -  .
0xabcd - 16   
0d9999 - 10    -   ,  .
07777 - 8    
0b1111 -    
0o777 -  ,  
9999. - 10     -  
7777 - 8      -  
*/
bool AdvancedNumberParser(int &result, wchar_t &ch)
{
	int nTmp = 0;
	bool bRet = true;
	std::wstring str;

	if (ch == L'0') //    0
	{
		//      
		if (g_pReader->getNextChar() == L'x' || g_pReader->getNextChar() == L'X')
		{
			ch = g_pReader->readChar();
			ch = g_pReader->readChar();
			//  16 
			bRet = HexNumberParser(ch, nTmp);
			goto l_aex;
		}
		else if (g_pReader->getNextChar() == L'd' || g_pReader->getNextChar() == L'D')
		{
			ch = g_pReader->readChar();
			ch = g_pReader->readChar();
			//  10 
			bRet = DecNumberParser(ch, nTmp);
			goto l_aex;
		}
		else if (g_pReader->getNextChar() == L'b' || g_pReader->getNextChar() == L'B')
		{
			ch = g_pReader->readChar();
			ch = g_pReader->readChar();
			//  2 
			bRet = BinNumberParser(ch, nTmp);
			goto l_aex;
		}
	}

	//  ,   , 8  10,     .
	//  0123. ,    ,    0
	bool bHas8 = false;

	while (g_pReader->getCurrCharType() == CReader::CHAR_TYPE::DIGITS)
	{
		str.push_back(ch); //     .

		//  ,      8  9
		if (ch == L'8' || ch == L'9')
		{
			bHas8 = true;
		}

		ch = g_pReader->readChar();
	}

	if (ch == L'.') //    
	{
		ch = g_pReader->readChar(true); //   
		//  10 
		nTmp = wcstol(str.c_str(), nullptr, 10);
	}
	else if (bHas8)
	{
		// ,    
		nTmp = 0;
		bRet = false;
	}
	else
	{
		//  8 
		nTmp = wcstol(str.c_str(), nullptr, 8);
	}

l_aex:
	result = nTmp;

	if (!bRet)
	{
		OutError(ERROR_135); //   
	}

	if (nTmp > 65535)
	{
		result &= 0xffff;
		OutError(ERROR_134); //  
		return false;
	}

	return bRet;
}

//        ^,
//       ,
//  : ch -   ^ .
/*
  11:
^xabcd - 16   
^habcd - 16   
^d9999 - 10   
^o7777 - 8    
^b1111 -    
^f12.3 -  .
^rabc -     Radix - 50
^c -     
*/
bool Macro11NumberParser(int &result, wchar_t &ch, bool &bInvert)
{
	int nTmp = 0;
	ch = g_pReader->readChar();
	bool bRet = true;

	switch (ch)
	{
		case L'X':
		case L'H':
			ch = g_pReader->readChar();
			//  16 
			bRet = HexNumberParser(ch, nTmp);
			break;

		case L'D':
			ch = g_pReader->readChar();
			//  10 
			bRet = DecNumberParser(ch, nTmp);
			break;

		case L'O':
			ch = g_pReader->readChar();
			//  8 
			bRet = OctNumberParser(ch, nTmp);
			break;

		case L'B':
			ch = g_pReader->readChar();
			//  2 
			bRet = BinNumberParser(ch, nTmp);
			break;

		case L'F':
		{
			ch = g_pReader->readChar();
			uint16_t flt[2];

			//   
			if (bRet = parse_float(ch, 1, flt))
			{
				nTmp = flt[0];
			}

			break;
		}

		case L'R':
		{
			ch = g_pReader->readChar();
			//  
			result = 0;
			int nMultipler = 03100;

			for (int i = 0; i < 3; ++i) //    
			{
				int nChCode = 0;

				for (int n = 0; n < 050; ++n)
				{
					if (RADIX50[n] == ch)
					{
						nChCode = n;    // 
						break;
					}
				}

				//      .
				result += nChCode * nMultipler; //   
				nMultipler /= 050;
				//  
				ch = g_pReader->readChar();

				//  , -  
				//  CT_LN -   ,   
				//  CT_SPACES -   .        
				//  ; -   
				//  - 
				if (ch == L',' || ch == L';' || g_pReader->getCurrCharType() == CReader::CHAR_TYPE::LN /* || g_pReader->getCurrCharType() == CReader::CHAR_TYPE::SPACES*/
				        || (ch == L'/' && (g_pReader->getNextChar() == L'/' || g_pReader->getNextChar() == L'*')))
				{
					// 
					break;
				}
			}

			return true;
		}

		case L'C':
			ch = g_pReader->readChar();
			bInvert = true;
			g_GlobalParameters.bInvert = !g_GlobalParameters.bInvert; //  
			return true;

		default:
			result = 0; //  -    .
			return false;
	}

	result = nTmp;

	if (!bRet)
	{
		OutError(ERROR_135); //   
	}

	if (nTmp > 65535)
	{
		result &= 0xffff;
		OutError(ERROR_134); //  
		return false;
	}

	return bRet;
}
