#pragma once

//        128  255,  
static const wchar_t koi8tbl[128] =   // {200..237}     .
{
	L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',
	L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',
	L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',
	L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',   L' ',
	// {240..277}
	0xb6,   0x2534, 0x2665, 0x2510, 0x2561, 0x251c, 0x2514, 0x2550,
	0x2564, 0x2660, 0x250c, 0x252c, 0x2568, 0x2193, 0x253c, 0x2551,
	0x2524, 0x2190, 0x256c, 0x2191, 0x2663, 0x2500, 0x256b, 0x2502,
	0x2666, 0x2518, 0x256a, 0x2565, 0x2567, 0x255e, 0x2192, 0x2593,
	// {300..337}
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	// {340..377}
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L'',
	L'',   L'',   L'',   L'',   L'',   L'',   L'',   L''
};


class CBKToken;

class CReader
{
	public:
		enum class FILE_CHARSET
		{
			FILEERROR,
			UNDEFINE,
			KOI8,
			CP866,
			CP1251,
			UTF8,
			UTF16LE
		};
		enum class CHAR_TYPE
		{
			CT_EOF,     //  
			LN,         //  
			SPACES,     //  
			LETTERS,    // 
			DIGITS,     // 
			OTHERS      //  
		};

	private:
		std::wstring m_strFileName;         //   
		wchar_t     m_chPrev;
		wchar_t     m_chCurrent;            //  
		wchar_t     m_chNext;               //  
		CHAR_TYPE   m_nCurrCharType;        //   
		FILE       *m_file;
		FILE       *m_dfile;
		FILE_CHARSET m_nCharset;
		int         m_nLineNum;             //   
		std::wstring m_strCur;              //   .       .
		size_t      m_nLen;

		/*
		      (ALT, WIN, KOI)
		 : http://ivr.webzone.ru/articles/defcod_2/
		(c)  , , 2004.
		*/

		static uint8_t table_2s[128];
		/* =========================================================================
		  alt2num.
		: a -      ALT.
		:     (0-31).
		========================================================================= */
		int alt2num(int a)
		{
			if (a >= 0xE0)
			{
				a -= 0x30;
			}

			return (a & 31);
		}
		/* =========================================================================
		  koi2num.
		: a -      KOI.
		:     (0-31).
		========================================================================= */

		int koi2num(int a)
		{
			static uint8_t t[32] =
			{
				30, 0, 1, 22, 4, 5, 20, 3, 21, 8, 9, 10, 11, 12, 13, 14, 15, 31,
				16, 17, 18, 19, 6, 2, 28, 27, 7, 24, 29, 25, 23, 26
			};
			return (t[a & 31]);
		}

		/* =========================================================================
		  work_2s -   .
		:  1 -     (0-31),
		c2 -     (0-31),
		check -   ,    
		(1 - , 0 - ),
		buf -       .
		: 0 -     ,
		1 -       ,
		2 -       .
		========================================================================= */
		int work_2s(int c1, int c2, int check, uint8_t buf[128]);

		/* =========================================================================
		  def_code -   .
		:
		n -      (1-255), 
		   .
		:  
		========================================================================= */
		FILE_CHARSET    def_code(int n = 128);
		int             get_char();
		//     .
		wchar_t         getChar();
		FILE_CHARSET    AnalyseCharset(const std::wstring &strFileName);

		void            getString();

	public:
		CReader(const std::wstring &strFileName, FILE_CHARSET nInputCharset = FILE_CHARSET::UNDEFINE);
		virtual ~CReader();

		const std::wstring &GetFileName()
		{
			return m_strFileName;
		}

		FILE_CHARSET GetFileCharset()
		{
			return m_nCharset;
		}

		wchar_t readChar(bool bSkipWS = false);

		/*
		*  ,   
		*/
		inline wchar_t getCurrChar()
		{
			return m_chCurrent;
		}

		/*
		*   ,   
		*/
		inline wchar_t getNextChar()
		{
			return m_chNext;
		}

		/*
		*    (  ).
		*/
		inline wchar_t getPrevChar()
		{
			return m_chPrev;
		}

		inline CHAR_TYPE getCurrCharType()
		{
			return m_nCurrCharType;
		}

		inline CHAR_TYPE getNextCharType()
		{
			return AnalyseChar(m_chNext);
		}

		bool isEOF();
		CHAR_TYPE AnalyseChar(wchar_t ch);
		bool SkipWhitespaces(wchar_t &ch);
		bool readToken(CBKToken *token, wchar_t &ch);

		inline const std::wstring &getCurString()
		{
			return m_strCur;
		}

		inline int getLineNum()
		{
			return m_nLineNum;
		}
};

void UNICODEtoBK(std::wstring &ustr, uint8_t *pBuff, size_t bufSize, bool bFillBuf);
uint8_t UNICODEtoBK_ch(wchar_t uchr);

/*
  ,   m_chCurrent -  ,
m_chNext -   ,      
  .
 */
